%slope
% Interactive GUI which shows forces and acceleration acting on a point
% particle being towed (via a light inextensible rope) up a rough inclined
% plane.
%
% LAST UPDATED by Andy French January 2017

function varargout = slope(varargin)

%% Initialization code - DO NOT EDIT
gui_Singleton = 1;
gui_State = struct('gui_Name',       mfilename, ...
    'gui_Singleton',  gui_Singleton, ...
    'gui_OpeningFcn', @slope_OpeningFcn, ...
    'gui_OutputFcn',  @slope_OutputFcn, ...
    'gui_LayoutFcn',  [] , ...
    'gui_Callback',   []);
if nargin && ischar(varargin{1})
    gui_State.gui_Callback = str2func(varargin{1});
end

if nargout
    [varargout{1:nargout}] = gui_mainfcn(gui_State, varargin{:});
else
    gui_mainfcn(gui_State, varargin{:});
end

%%

%% slope opening function. This executes just before slope is made visible.
function slope_OpeningFcn(hObject, eventdata, handles, varargin)

% Choose default command line output for slope
handles.output = hObject;

%Turn off warnings (e.g. 'divide by zero')
warning off

%Set up global variable d which can be accessed by all functions
global d
d.h = handles;
d.legend = [];

%Center figure
figcentre( d.h.FIGUREslope );

%Load 'slidy object' image
d.I = imread('sybil.jpg');
d.I = double(flipdim(d.I,1))/255;
dim = size(d.I);
d.Iwidth = dim(2);
d.Iheight = dim(1);

%Set defaults
slope_defaults

%Update slope
slope_update;

%%

% Default parameters
function slope_defaults
global d

%Plot acceleration?
d.plot_acc = 0;

%Font size for graphs
d.fsize = 18;

%Arrow size for vector diagrams
d.arrowsize = 4;

%Arrow RGB colours
d.arrow_colours{1} = [1,0,0];
d.arrow_colours{2} = [0,1,0];
d.arrow_colours{3} = [0,0,1];
d.arrow_colours{4} = [1,0,1];
d.arrow_colours{5} = [0,0,0];

%Default parameters
d.mu = 0.3;
d.m = 4;
d.theta = 30;
d.phi = 45;
d.g = 9.81;
d.T = 10;

%Set changes in parameters when + or - buttons pressed
d.dmu = 0.01;
d.dm = 0.1;
d.dtheta = 1;
d.dphi = 1;
d.dg = 0.1;
d.dT = 0.2;

%Store defaults
d.mu_default = d.mu;
d.m_default = d.m;
d.theta_default = d.theta;
d.phi_default = d.phi;
d.g_default = d.g;
d.T_default = d.T;

%Set GUI with defaults
set( d.h.EDITmu,'string',num2str(d.mu) );
set( d.h.EDITmass,'string',num2str(d.m) );
set( d.h.EDITslopeangle,'string',num2str(d.theta) );
set( d.h.EDITropeangle,'string',num2str(d.phi) );
set( d.h.EDITg,'string',num2str(d.g) );
set( d.h.EDITT,'string',num2str(d.T) );
set( d.h.AXESfreebody,'fontsize',d.fsize );
set( d.h.AXESvectoradd,'fontsize',d.fsize );

%Initialize various GUI flags
%(don't change these, as they need to start as being zero)
d.plot_and_save_PNG = 0;
d.fix_scale = 0;

%Set width of image (that is sliding down the slope!)
d.w = 10;

%%

%% Update GUI
function slope_update
global d

%Convert angles into radians
t = d.theta*pi/180;
p = d.phi*pi/180;

%Normal contact force
R = d.m*d.g*cos(t) - d.T*sin(p);

%Compute vector components of R and T
Rx = -R*sin(t);
Ry = R*cos(t);
Tx = d.T*cos(p+t);
Ty = d.T*sin(p+t);

%Determine scenario, based upon vlaue of  T. Calculate acceleration a
%and friction force F, plus vectorvcomponents
aup = NaN;

%Does friction force act uphill or downhill?
if d.T*cos(p) > d.m*d.g*sin(t)
    
    %F acts downhill
    F = d.T*cos(p) - d.m*d.g*sin(t);
    
    %Is mass in equilibrium?
    if F <= d.mu*R
        %Yes. No acceleration
        a = 0;
        ax = 0;
        ay = 0;
        set( d.h.TEXTdescription,'string','Static, friction acts downhill' );
        set( d.h.TEXTacc,'string','a = 0 ms/^2' );
    else
        %Acceleration is uphill
        aup = 1;
        F = d.mu*R;
        a = d.T*cos(p)/d.m - F/d.m - d.g*sin(t);
        ax = a*cos(t);
        ay = a*sin(t);
        set( d.h.TEXTdescription,'string','Acceleration uphill' );
        set( d.h.TEXTacc,'string',['a = ',num2str(a,3),' m/s^2'] );
    end
    Fx = -F*cos(t);
    Fy = -F*sin(t);
    
else
    
    %F acts uphill
    F = -d.T*cos(p) + d.m*d.g*sin(t);
    
    %Is mass in equilibrium?
    if F <= d.mu*R
        %Yes. No acceleration
        a = 0;
        ax = 0;
        ay = 0;
        set( d.h.TEXTdescription,'string','Static, friction  acts uphill' );
        set( d.h.TEXTacc,'string','a = 0 ms/^2' );
    else
        %Acceleration is downhill
        aup = -1;
        F = d.mu*R;
        a = -d.T*cos(p)/d.m - F/d.m + d.g*sin(t);
        ax = -a*cos(t);
        ay = -a*sin(t);
        set( d.h.TEXTdescription,'string','Acceleration downhill' );
        set( d.h.TEXTacc,'string',['a = ',num2str(a,3),' m/s^2'] );
    end
    Fx = F*cos(t);
    Fy = F*sin(t);
    
end

%Set force vector arrowhead size to be a fraction of the meadian force
arrowheadsize = 0.03*mean([d.m*d.g,F,R,d.T,d.m*a]);

%Perform zero checks on vectors
c1 = notzerovec([Rx,Ry]);
c2 = notzerovec([Tx,Ty]);
c3 = notzerovec([Fx,Fy]);
c4 = notzerovec([0,-d.m*d.g]);
c5 = notzerovec(d.m*[ax,ay]);

%Plot free body diagram
if d.plot_and_save_PNG == 0;
    axes( d.h.AXESfreebody );
    cla;
else
    fig = figure('name','slope','color',[1 1 1]);
end
hold on;

%Plot image
x = linspace(-d.w,d.w,d.Iwidth);
y = linspace(0,2*d.w*d.Iheight/d.Iwidth,d.Iheight);
[x,y] = meshgrid(x,y);
dim = size(x);
xy = [cos(t),-sin(t);sin(t),cos(t)]*[reshape(x,[1,numel(x)]) ; reshape(y,[1,numel(y)]) ];
x = xy(1,:);
y = xy(2,:);
x = reshape(x,dim);
y = reshape(y,dim);
clear xy;
surf(x,y,ones(dim),d.I);
shading interp

%Determine start coordinates of vectors, so they are not obscured by Sybil!
xhat = [cos(t);sin(t)];
yhat = [-sin(t);cos(t)];

%Rope tension
r = d.w*xhat + d.w*d.Iheight/d.Iwidth*yhat;
Tx0 = r(1); Ty0 = r(2);

%Reaction force
r =  2*d.w*d.Iheight/d.Iwidth*yhat;
Rx0 = r(1); Ry0 = r(2);

%Friction
if d.T*cos(p) > d.m*d.g*sin(t)
    %Downhill
    r = -d.w*xhat;
else
    %Uphill
    r = d.w*xhat;
end
Fx0 = r(1); Fy0 = r(2);

%Mass x acceleration
if aup ==1
    r = d.w*xhat + d.w*d.Iheight/d.Iwidth*yhat;
    max0 = r(1); may0 = r(2);
elseif aup == -1
    r = -d.w*xhat + d.w*d.Iheight/d.Iwidth*yhat;
    max0 = r(1); may0 = r(2);
end

%Plot force vectors
arrowheadsize = 2*arrowheadsize;
plot( [Rx0,Rx0+Rx],[Ry0,Ry+Ry0],'linewidth',d.arrowsize,'color',d.arrow_colours{1} );
if c1==1; drawarrow( Rx0, Rx0+Rx, Ry0, Ry0+Ry, arrowheadsize, d.arrow_colours{1} ); end
plot( [Tx0,Tx0+Tx],[Ty0,Ty0+Ty],'linewidth',d.arrowsize,'color',d.arrow_colours{2} );
if c2==1; drawarrow( Tx0, Tx0+Tx, Ty0, Ty0+Ty, arrowheadsize, d.arrow_colours{2} ); end
plot( [Fx0,Fx0+Fx],[Fy0,Fy0+Fy],'linewidth',d.arrowsize,'color',d.arrow_colours{3} );
if c3==1; drawarrow( Fx0, Fx0+Fx, Fy0, Fy0+Fy, arrowheadsize, d.arrow_colours{3} ); end
plot( [0,0],[0,-d.m*d.g],'linewidth',d.arrowsize,'color',d.arrow_colours{4} );
if c4==1; drawarrow( 0, 0, 0, -d.m*d.g, arrowheadsize, d.arrow_colours{4} ); end
if d.plot_acc == 1
    if isnan(aup)==0
        plot( [max0,max0+d.m*ax],[may0,may0+d.m*ay],'linewidth',d.arrowsize,'color',d.arrow_colours{5} );
        drawarrow( max0,max0+d.m*ax,may0,may0+d.m*ay, arrowheadsize, d.arrow_colours{5} );
    end
end

%Show components of T and W
T1x = Tx0+d.T*cos(t)*cos(p);
T1y = Ty0+d.T*cos(p)*sin(t);
T2x = Tx0+d.T*cos(t+p);
T2y = Ty0+d.T*sin(t+p);
plot( [Tx0,T1x,T2x],[Ty0,T1y,T2y],'g--' );
Wx = d.m*d.g*cos(t)*sin(t);
Wy = -d.m*d.g*cos(t)*cos(t);
plot( [0,Wx,0],[0,Wy,-d.m*d.g],'m--' );

%Plot labels
text( Rx0+Rx - 2*arrowheadsize, Ry0+Ry + 4*arrowheadsize,['R = ',num2str(R,3)],'fontsize',d.fsize );
text( Tx0+Tx - 2*arrowheadsize, Ty0+Ty + 4*arrowheadsize,['T = ',num2str(d.T,3)],'fontsize',d.fsize );
if d.T*cos(p) > d.m*d.g*sin(t)
    %Friction acts downhill
    text( Fx0+Fx - 3*arrowheadsize, Fy0+Fy + 4*arrowheadsize,['F = ',num2str(F,3)],...
        'fontsize',d.fsize,'HorizontalAlignment','right' );
else
    text( Fx0+Fx + arrowheadsize, Fy0+Fy + 4*arrowheadsize,['F = ',num2str(F,3)],...
        'fontsize',d.fsize,'HorizontalAlignment','left'  );
end
text( 0, -d.m*d.g - 4*arrowheadsize,['W = ',num2str(d.m*d.g,3)],'fontsize',d.fsize );
if d.plot_acc == 1
    if isnan(aup)==0
        if aup==1
            text( max0+d.m*ax + 2*arrowheadsize, may0+d.m*ay + 4*arrowheadsize,['ma = ',num2str(d.m*a,3)],'fontsize',d.fsize );
        else
            text( max0+d.m*ax - 2*arrowheadsize, may0+d.m*ay - 4*arrowheadsize,['ma = ',num2str(d.m*a,3)],...
                'fontsize',d.fsize,'HorizontalAlignment','right' );
        end
    end
end

%Axis scale
axis equal
if d.fix_scale == 0
    if d.plot_and_save_PNG == 0;
        
        %Tightly enclose vectors with axis
        axis tight
        d.xlimits_freebody = get(d.h.AXESfreebody,'xlim');
        d.ylimits_freebody = get(d.h.AXESfreebody,'ylim');
        
        %Enhance xlimits to make slope look more significant
        d.xlimits_freebody = 3*d.xlimits_freebody;
        set( d.h.AXESfreebody,'xlim',d.xlimits_freebody,'ylim',d.ylimits_freebody );
    else
        %Set axis in new figure to have the same limits as the GUI axis
        set( gca,'xlim',d.xlimits_freebody,'ylim',d.ylimits_freebody );
    end
else
    set( d.h.AXESfreebody,'xlim',d.xlimits_freebody );
    set( d.h.AXESfreebody,'ylim',d.ylimits_freebody );
end
grid on
box on

%Plot slope
x1 = d.xlimits_freebody(1);
x2 = d.xlimits_freebody(2);
x3 = d.xlimits_freebody(2);
y1 = d.xlimits_freebody(1)*tan(t);
y2 = d.xlimits_freebody(1)*tan(t);
y3 = d.xlimits_freebody(2)*tan(t);
patch( [x1,x2,x3],[y1,y2,y3],[-1,-1,-1],[0.8,0.8,0.8],'edgecolor','none' );
axis off

%Print PNG file?
if d.plot_and_save_PNG == 1;
    filename = strrep( datestr(now),':','-' );
    filename = ['free body ',filename,'.png'];
    %print( fig, filename, '-dpng','-r300' );
    imwrite( frame2im(getframe(gcf)), filename, 'PNG');
end

%%

%Plot vector addition diagram
if d.plot_and_save_PNG == 0;
    axes( d.h.AXESvectoradd );
    cla;
else
    clf;
end
hold on;
x1 = Rx;
x2 = Rx+Tx;
x3 = Rx+Tx+Fx;
x4 = Rx+Tx+Fx;
y1 = Ry;
y2 = Ry+Ty;
y3 = Ry+Ty+Fy;
y4 = Ry+Ty+Fy - d.m*d.g;
plot( [0,x1],[0,y1],'linewidth',d.arrowsize,'color',d.arrow_colours{1} );
plot( [x1,x2],[y1,y2],...
    'linewidth',d.arrowsize,'color',d.arrow_colours{2} );
plot( [x2,x3],[y2,y3],...
    'linewidth',d.arrowsize,'color',d.arrow_colours{3} );
plot( [x3,x4],[y3,y4],...
    'linewidth',d.arrowsize,'color',d.arrow_colours{4} );

%Plot arrows
arrowheadsize = 0.5*arrowheadsize;
if c1==1; drawarrow( 0, x1, 0, y1, arrowheadsize, d.arrow_colours{1} ); end
if c2==1; drawarrow( x1, x2, y1, y2, arrowheadsize, d.arrow_colours{2} ); end
if c3==1; drawarrow( x2, x3, y2, y3, arrowheadsize, d.arrow_colours{3} ); end
if c4==1; drawarrow( x3, x4, y3, y4, arrowheadsize, d.arrow_colours{4} ); end

%Plot resultant mass x acceleration
if d.plot_acc == 1
    plot( [0,d.m*ax],[0,d.m*ay],'linewidth',d.arrowsize,'color',d.arrow_colours{5} );
    if c5==1; drawarrow( 0,d.m*ax,0,d.m*ay, arrowheadsize, d.arrow_colours{5} ); end
end

%Axis scale
axis equal
axis tight
d.xlimits_vectoradd = get(d.h.AXESvectoradd,'xlim');
d.ylimits_vectoradd = get(d.h.AXESvectoradd,'ylim');
grid on
box on

%Print PNG file?
if d.plot_and_save_PNG == 1;
    filename = strrep( datestr(now),':','-' );
    filename = ['vector add ',filename,'.png'];
    print( fig, filename, '-dpng','-r300' );
    close(fig);
    figure(d.h.FIGUREslope);
end

%%

%% Diffractor output function. Outputs from this function are returned to the command line.
function varargout = slope_OutputFcn(hObject, eventdata, handles)
global d

% Get default command line output from handles structure
varargout{1} = d.h.output;

%%

%% Toggle fix axes scale for plot
function TOGGLEfixaxesscale_Callback(hObject, eventdata, handles)
global d
if d.fix_scale==1
    d.fix_scale = 0;
else
    d.fix_scale = 1;
end
slope_update

%%

%% Save PNG of graph
function PUSHsavepng_Callback(hObject, eventdata, handles)
global d

%Force slope_update to generate figures in a new window
d.plot_and_save_PNG = 1;
slope_update;
d.plot_and_save_PNG = 0;

%%

%% Hide vector addition graph
function CHECKhide_Callback(hObject, eventdata, handles)
global d
if get(d.h.CHECKhide,'value') == 1
    set( d.h.TEXTbiggraybox,'visible','on' );
else
    set( d.h.TEXTbiggraybox,'visible','off' );
end

%%

%% Show acceleration
function TOGGLEshowacc_Callback(hObject, eventdata, handles)
global d
if d.plot_acc==1
    d.plot_acc = 0;
else
    d.plot_acc = 1;
end
slope_update

%%

%% Restore default parameters
function PUSHdefault_Callback(hObject, eventdata, handles)
global d

%Store defaults
d.mu = d.mu_default;
d.m = d.m_default;
d.theta = d.theta_default;
d.phi = d.phi_default;
d.g = d.g_default;
d.T = d.T_default;

%Set GUI with defaults
set( d.h.EDITmu,'string',num2str(d.mu) );
set( d.h.EDITmass,'string',num2str(d.m) );
set( d.h.EDITslopeangle,'string',num2str(d.theta) );
set( d.h.EDITropeangle,'string',num2str(d.phi) );
set( d.h.EDITg,'string',num2str(d.g) );
set( d.h.EDITT,'string',num2str(d.T) );
set( d.h.AXESfreebody,'fontsize',d.fsize );
set( d.h.AXESvectoradd,'fontsize',d.fsize );
slope_update;

%%

%% Coefficient of friction edit box
function EDITmu_Callback(hObject, eventdata, handles)
global d
d.mu = str2num( get( d.h.EDITmu,'string') );
if d.mu < 0
    d.mu = 0;
    set( d.h.EDITmu,'string','0');
end
slope_update;

%%

%% Coefficient of friction edit box + pushbutton
function PUSHmuplus_Callback(hObject, eventdata, handles)
global d
d.mu = d.mu + d.dmu;
set( d.h.EDITmu,'string',num2str(d.mu));
slope_update;

%%

%% Coefficient of friction edit box - pushbutton
function PUSHmuminus_Callback(hObject, eventdata, handles)
global d
d.mu = d.mu - d.dmu;
if d.mu < 0
    d.mu = 0;
end
set( d.h.EDITmu,'string',num2str(d.mu));
slope_update;

%%

%% Slope angle edit box
function EDITslopeangle_Callback(hObject, eventdata, handles)
global d
d.theta = str2num( get( d.h.EDITslopeangle,'string') );
if d.theta < 0
    d.theta = 0;
    set( d.h.EDITslopeangle,'string','0');
elseif d.theta > 90
    d.theta = 90;
    set( d.h.EDITslopeangle,'string','90');
end
slope_update;

%%

%% Slope angle edit box + pushbutton
function PUSHslopeangleplus_Callback(hObject, eventdata, handles)
global d
d.theta = d.theta + d.dtheta;
if d.theta > 90
    d.theta = 90;
end
set( d.h.EDITslopeangle,'string',num2str(d.theta));
slope_update;

%%

%% Slope angle edit box - pushbutton
function PUSHslopeangleminus_Callback(hObject, eventdata, handles)
global d
d.theta = d.theta - d.dtheta;
if d.theta < 0
    d.theta = 0;
end
set( d.h.EDITslopeangle,'string',num2str(d.theta));
slope_update;

%%

%% Rope angle edit box
function EDITropeangle_Callback(hObject, eventdata, handles)
global d
d.phi = str2num( get( d.h.EDITropeangle,'string') );
if d.phi < 0
    d.phi = 0;
    set( d.h.EDITropeangle,'string','0');
elseif d.phi > 90
    d.phi = 90;
    set( d.h.EDITropeangle,'string','90');
end
slope_update;

%%

%% Rope angle edit box + pushbutton
function PUSHropeangleplus_Callback(hObject, eventdata, handles)
global d
d.phi = d.phi + d.dphi;
if d.phi > 90
    d.phi = 90;
end
set( d.h.EDITropeangle,'string',num2str(d.phi));
slope_update;

%%

%% Rope angle edit box - pushbutton
function PUSHropeangleminus_Callback(hObject, eventdata, handles)
global d
d.phi = d.phi - d.dphi;
if d.phi < 0
    d.phi = 0;
end
set( d.h.EDITropeangle,'string',num2str(d.phi));
slope_update;

%%

%% Strength of gravity edit box
function EDITg_Callback(hObject, eventdata, handles)
global d
d.g = str2num( get( d.h.EDITg,'string') );
if d.g < 0
    d.g = 0;
    set( d.h.EDITg,'string','0');
end
slope_update;

%%

%% Strength of gravity edit box + pushbutton
function PUSHgplus_Callback(hObject, eventdata, handles)
global d
d.g = d.g + d.dg;
set( d.h.EDITg,'string',num2str(d.g));
slope_update;

%%

%% Strength of gravity edit box - pushbutton
function PUSHgminus_Callback(hObject, eventdata, handles)
global d
d.g = d.g - d.dg;
if d.g < 0
    d.g = 0;
end
set( d.h.EDITg,'string',num2str(d.g));
slope_update;

%%

%% Mass edit box
function EDITmass_Callback(hObject, eventdata, handles)
global d
d.m = str2num( get( d.h.EDITmass,'string') );
if d.m < 0
    d.m = 0;
    set( d.h.EDITmass,'string','0');
end
slope_update;

%%

%% Mass edit box + pushbutton
function PUSHmassplus_Callback(hObject, eventdata, handles)
global d
d.m = d.m + d.dm;
set( d.h.EDITmass,'string',num2str(d.m));
slope_update;

%%

%% Mass edit box - pushbutton
function PUSHmassminus_Callback(hObject, eventdata, handles)
global d
d.m = d.m - d.dm;
if d.m < 0
    d.m = 0;
end
set( d.h.EDITmass,'string',num2str(d.m));
slope_update;

%%

%% Rope tension edit box
function EDITT_Callback(hObject, eventdata, handles)
global d
d.T = str2num( get( d.h.EDITT,'string') );
Tmax = d.m*d.g*cos(d.theta*pi/180)/sin(d.phi*pi/180);
if d.T < 0
    d.T = 0;
    set( d.h.EDITT,'string','0');
elseif d.T > Tmax
    d.T = Tmax;
    set( d.h.EDITT,'string',num2str(Tmax));
end
slope_update;

%%

%% Rope tension edit box + pushbutton
function PUSHTplus_Callback(hObject, eventdata, handles)
global d
d.T = d.T + d.dT;
Tmax = d.m*d.g*cos(d.theta*pi/180)/sin(d.phi*pi/180);
if d.T > Tmax
    d.T = Tmax;
end
set( d.h.EDITT,'string',num2str(d.T,3));
slope_update;

%%

%% Rope tension edit box - pushbutton
function PUSHTminus_Callback(hObject, eventdata, handles)
global d
d.T = d.T - d.dT;
if d.T < 0
    d.T = 0;
end
set( d.h.EDITT,'string',num2str(d.T));
slope_update;

%%

%figcenter
% Centres a figure with handle fig in the current display.
function figcentre(fig)

%Get screen dimensions in pixels.
set(0,'units','pixels') ;
screen_size=get(0,'Screensize');

%Get original figure units
orig_units = get(fig,'units');

%Get position vector for figure in pixels
set(fig,'units','pixels');
fig_position=get(gcf,'position') ;

%Define coordinates or figure bottom hand corner from screen bottom hand corner
%that will place the figure window in the center of the screen.
centre_x = 0.5*screen_size(3) - 0.5*fig_position(3);
centre_y = 0.5*screen_size(4) - 0.5*fig_position(4);

%Modify figure position to centre
set(fig,'position',[centre_x,centre_y,fig_position(3),fig_position(4)] );

%Reset figure units
set(fig,'units',orig_units);

%%

%drawarrow
% Function which draws an arrow in the middle of a line defined by start
% and end coordinates [x0,x1] and [y0,y1]
function a = drawarrow( x0, x1, y0, y1, arrowsize, arrow_RGB )

%Define centre of arrowhead
xc = (x0+x1)/2;
yc = (y0+y1)/2;

%Rotation angle of arrow from horizontal in order for arrow to point along
%line
t = atan2( y1-y0,x1-x0 );

%Transform arrow coordinates
a = arrowsize;
r = [ 0, 2*a, a; 0, 0, 2*a];
r = r - [a,a,a;a,a,a];
r = [ sin(t), cos(t); -cos(t), sin(t)] * r ;
r = r + [xc,xc,xc; yc,yc,yc];
a = patch(r(1,:),r(2,:),arrow_RGB);
set(a,'Edgecolor',arrow_RGB );

%%

%notzerovec
% Function which returns a 1 if the vector x is not [0,0]
function y = notzerovec(x)
if sum(abs(x))==0
    y = 0;
else
    y = 1;
end

%End of code






